from enum import Enum


class Action(Enum):
    Interpretation = "Intrepretation"
    Attachment = "Attachment"
    Preparation = "Preparation"
    Run = "Run"
    Stop = "Stop"
    Installation = "Installation"
    Removal = "Removal"
    Results = "Results"
