{% if args %}
{% for arg in args -%}
{{ arg[0] }}: {{ arg[1] or "Any"}},
{% endfor -%}
{%- endif -%}
{% if sink -%}
{%- if args|length > 0 -%}
/,
*{{sink[0]}}: Any,
{%- else -%}
*{{sink[0]}}: Any,
{%- endif -%}
{%- else -%}
{%- if args|length > 0 -%}
/,
{%- else -%}
*,
{%- endif -%}
{%- endif %}
{%- if config_fields %}
{%- for key, field in config_fields.items() %}
{{  key }}: Annotated["{{ field.annotation }}", ConfigArgument(required={{field.is_required()}})] = None,
{%- endfor %}
{%- endif %}
{% for rkey, rtype in reserved_config_fields -%}
{{rkey}}: Optional["{{rtype}}"] = None,
{%- endfor %}
{%- if config_model %}
config: Optional["{{config_model}}"] = None,
{%- else %}
config: Optional[dict] = None,
{% endif %}
**kwargs: Any